import argparse
import sys
from io import BytesIO

from PIL import Image
from PIL.ImageOps import scale as scale_image
from rich.console import Console
from rich.table import Table

from . import iterm
from . import kitty
from . import sixel
from . import text
from .terminal import get_info
from .terminal import get_preferred_graphics_protocol
from .utils import pad_ratio


def _do_show(args):
    maximum_size_pixels = None

    if args.size is not None:
        size = tuple([int(v) for v in args.size.split('x')])
    else:
        size = None
        term_size = get_info().size

        if term_size.pixels is not None:
            maximum_size_pixels = (
                (term_size.cells[0] - 1) * term_size.cell_pixels[0],
                (term_size.cells[1] - 2) * term_size.cell_pixels[1])

    for file in args.files:
        if file == '-':
            image = Image.open(BytesIO(sys.stdin.buffer.read()))
        else:
            image = Image.open(file)

        if maximum_size_pixels is not None:
            scale = min(maximum_size_pixels[0] / image.width,
                        maximum_size_pixels[1] / image.height)

            if scale < 1:
                image = scale_image(image, scale)

        write(image,
              sys.stdout.buffer,
              size,
              True,
              args.protocol)
        sys.stdout.buffer.write(b'\n')


def _yes_or_no(value):
    if value:
        return 'yes'
    else:
        return 'no'


def _do_info(args):
    info = get_info()
    table = Table('Name', 'Value')

    table.add_row('Rows', str(info.size.cells[1]))
    table.add_row('Columns',    str(info.size.cells[0]))

    if info.size.pixels is not None:
        table.add_row('Width',      str(info.size.pixels[0]))
        table.add_row('Height',     str(info.size.pixels[1]))

    if info.size.cell_pixels is not None:
        table.add_row('CellWidth',  str(info.size.cell_pixels[0]))
        table.add_row('CellHeight', str(info.size.cell_pixels[1]))

    sixel = info.graphics.sixel
    table.add_row('SixelSupport', _yes_or_no(sixel.is_supported))
    kitty = info.graphics.kitty
    table.add_row('KittySupport', _yes_or_no(kitty.is_supported))

    if kitty.is_supported:
        transmission_mediums = ', '.join(kitty.transmission_mediums)
        table.add_row('KittyTransmissionMediums', transmission_mediums)

    iterm = info.graphics.iterm
    table.add_row('ITermSupport', _yes_or_no(iterm.is_supported))

    Console().print(table)


def _main():
    parser = argparse.ArgumentParser()

    # Workaround to make the subparser required in Python 3.
    subparsers = parser.add_subparsers(title='subcommands',
                                       dest='subcommand')
    subparsers.required = True

    subparser = subparsers.add_parser('show')
    subparser.add_argument('--size', help='Width and height in cells.')
    subparser.add_argument('--protocol',
                           help='Protocol to use.')
    subparser.add_argument('files', nargs="+")
    subparser.set_defaults(func=_do_show)

    subparser = subparsers.add_parser('info')
    subparser.set_defaults(func=_do_info)

    args = parser.parse_args()

    try:
        args.func(args)
    except BaseException as e:
        sys.exit('error: ' + str(e))


def write(image,
          fout=None,
          size=None,
          move_cursor=True,
          protocol=None):
    """Write given image to given output stream, or stdout by default.

    Size is a tuple of width and height in cells.

    Protocol is the terminal protocol to use, for example 'kitty'.

    """

    if protocol is None:
        protocol = get_preferred_graphics_protocol()

    if fout is None:
        fout = sys.stdout.buffer

    if size is not None:
        image = pad_ratio(image, size, get_info().size.cell_pixels)

    if protocol == 'kitty':
        kitty.write(image, fout, size, move_cursor)
    elif protocol == 'iterm':
        png = BytesIO()
        image.save(png, 'png')
        iterm.write(png.getvalue(), fout, size)
    elif protocol == 'text':
        text.write(image, fout, size)
    else:
        raise Exception(f"Unsupported graphics protocol '{protocol}'.")


def write_file(path,
               fout=None,
               size=None,
               move_cursor=True,
               protocol=None):
    write(Image.open(path), fout, size, move_cursor, protocol)



import os
import unittest

from PIL import Image

from terminal_graphics import utils

from .utils import TestCase


class KittyTest(TestCase):

    def test_split_image_horizontally(self):
        image = Image.open('tests/files/test_split_image_horizontally/diagonal.png')
        images = list(utils.split_image_horizontally(image, 3))
        self.assertEqual(len(images), 3)
        os.makedirs('tests/build/test_split_image_horizontally', exist_ok=True)

        for i in range(3):
            actual = f'tests/build/test_split_image_horizontally/{i}.png'
            expected = f'tests/files/test_split_image_horizontally/{i}.png'
            images[i].save(actual)
            self.assert_files_equal(actual, expected)



import numpy

from . import ctext
from .terminal import get_size


def write(image, fout, size=None):
    if size is None:
        terminal_size = get_size()
        width_cells = terminal_size.cells[0] - 2
        height_cells = terminal_size.cells[1] - 2
        pixels_per_cell = max(image.width // width_cells,
                              image.height // (2 * height_cells))
        size = (min(width_cells, image.width // pixels_per_cell),
                min(height_cells, image.height // (2 * pixels_per_cell)))

    image = image.resize((size[0], 2 * size[1])).convert('RGB')
    ctext.write(image.height // 2, image.width, numpy.array(image))



from PIL.ImageOps import pad


def split_image_horizontally(image, number_of_images):
    """Split given image horizontally into given number of images.

    """

    height = (image.height + number_of_images - 1) // number_of_images

    for y in range(0, image.height, height):
        yield image.crop((0, y, image.width, y + height))


def pad_ratio(image, size, cell_size=None):
    """Pad to fit given size. Keeps aspect ratio.

    """

    if cell_size is None:
        cell_size = (1, 2)

    image_ratio = image.width / image.height
    size_ratio = (size[0] * cell_size[0]) / (size[1] * cell_size[1])

    if image_ratio < size_ratio:
        width = int(image.height * size_ratio)
        height = image.height
    else:
        width = image.width
        height = int(image.width / size_ratio)

    return pad(image, (width, height))



import unittest
from io import BytesIO

from terminal_graphics import kitty


class KittyTest(unittest.TestCase):

    def test_write_png(self):
        output = BytesIO()

        with open('tests/files/box.png', 'rb') as fin:
            kitty.write_png(fin.read(), output)

        self.assertEqual(
            output.getvalue(),
            b'\x1b_Gm=0,a=T,f=100;'
            b'iVBORw0KGgoAAAANSUhEUgAAAD4AAAA6CAYAAADoUOpSAAAACXBIWXMAAAPYAA'
            b'AD2AFuR2M1AAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAA'
            b'AF5JREFUaIHtz0EBACAQgDC1f5sLqDF8jCWAPTN3gc7vgF8a1zSuaVzTuKZxTe'
            b'OaxjWNaxrXNK5pXNO4pnFN45rGNY1rGtc0rmlc07imcU3jmsY1jWsa1zSuaVzT'
            b'uOYBiyUDjLZLEGgAAAAASUVORK5CYII='
            b'\x1b\\')

    def test_write_png_size(self):
        output = BytesIO()

        with open('tests/files/box.png', 'rb') as fin:
            kitty.write_png(fin.read(), output, (10, 5))

        self.assertEqual(
            output.getvalue(),
            b'\x1b_Gm=0,a=T,f=100,c=10,r=5;'
            b'iVBORw0KGgoAAAANSUhEUgAAAD4AAAA6CAYAAADoUOpSAAAACXBIWXMAAAPYAA'
            b'AD2AFuR2M1AAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAA'
            b'AF5JREFUaIHtz0EBACAQgDC1f5sLqDF8jCWAPTN3gc7vgF8a1zSuaVzTuKZxTe'
            b'OaxjWNaxrXNK5pXNO4pnFN45rGNY1rGtc0rmlc07imcU3jmsY1jWsa1zSuaVzT'
            b'uOYBiyUDjLZLEGgAAAAASUVORK5CYII='
            b'\x1b\\')



#
# https://sw.kovidgoyal.net/kitty/graphics-protocol/
#

from base64 import b64encode
from io import BytesIO

import numpy


def _serialize_gr_command(payload, **cmd):
    cmd = ','.join(f'{k}={v}' for k, v in cmd.items())
    data = []
    data.append(b'\x1b_G')
    data.append(cmd.encode('ascii'))

    if payload:
        data.append(b';')
        data.append(payload)

    data.append(b'\x1b\\')

    return b''.join(data)


def _write_chunked(fout, data, **cmd):
    data = b64encode(data)
    reader = BytesIO(data)
    more = 1

    while more == 1:
        chunk = reader.read(4096)
        more = 1 if reader.tell() < len(data) else 0
        fout.write(_serialize_gr_command(chunk, m=more, **cmd))
        cmd.clear()


def _write_rgb_rgba(data, width, height, fout, f, size, move_cursor):
    cmd = {}

    if not move_cursor:
        cmd['C'] = 1

    if size is not None:
        cmd['c'] = size[0]
        cmd['r'] = size[1]

    _write_chunked(fout, data, a='T', f=f, s=width, v=height, **cmd)


def write_rgb(data, width, height, fout, size=None, move_cursor=True):
    _write_rgb_rgba(data, width, height, fout, 24, size, move_cursor)


def write_rgba(data, width, height, fout, size=None, move_cursor=True):
    _write_rgb_rgba(data, width, height, fout, 32, size, move_cursor)


def write_png(data, fout, size=None, move_cursor=True):
    cmd = {}

    if not move_cursor:
        cmd['C'] = 1

    if size is not None:
        cmd['c'] = size[0]
        cmd['r'] = size[1]

    _write_chunked(fout, data, a='T', f=100, **cmd)


def write(image, fout, size=None, move_cursor=True):
    image = image.convert('RGBA')
    write_rgba(numpy.array(image), image.width, image.height, fout, size, move_cursor)


def delete_all_on_screen(fout):
    fout.write(b'\x1b_Ga=d\x1b\\')


def delete_all_intersecting_with_cursor(fout):
    fout.write(b'\x1b_Ga=d,d=c\x1b\\')


def delete_all_intersecting_with_cell(fout, row, column):
    fout.write(f'\x1b_Ga=d,d=p,x={column},y={row}\x1b\\'.encode('utf-8'))



import fcntl
import os
import struct
import termios
from dataclasses import dataclass
from typing import List
from typing import Tuple


@dataclass
class Size:
    cells: Tuple[int, int]
    pixels: Tuple[int, int] = None
    cell_pixels: Tuple[int, int] = None


@dataclass
class KittyGraphicsInfo:
    is_supported: bool
    transmission_mediums: List[str]


@dataclass
class SixelGraphicsInfo:
    is_supported: bool


@dataclass
class ITermGraphicsInfo:
    is_supported: bool


@dataclass
class GraphicsInfo:
    has_true_color: bool
    sixel: SixelGraphicsInfo
    kitty: KittyGraphicsInfo
    iterm: ITermGraphicsInfo


@dataclass
class Info:
    size: Size
    graphics: GraphicsInfo


def get_size():
    rows, columns, width, height = struct.unpack(
        'HHHH',
        fcntl.ioctl(1,
                    termios.TIOCGWINSZ,
                    struct.pack('HHHH', 0, 0, 0, 0)))

    size = Size((columns, rows))

    if width > 0 and height > 0:
        size.pixels = (width, height)
        size.cell_pixels = (width // columns, height // rows)

    return size


def get_graphics_info():
    return GraphicsInfo(os.environ.get('COLORTERM') == 'truecolor',
                        get_sixel_graphics_info(),
                        get_kitty_graphics_info(),
                        get_iterm_graphics_info())


def get_sixel_graphics_info():
    return SixelGraphicsInfo(False)


def get_kitty_graphics_info():
    term = os.environ.get('TERM')
    is_supported = term is not None and 'kitty' in term
    transmission_mediums = []

    if is_supported:
        transmission_mediums.append('direct')

    return KittyGraphicsInfo(is_supported, transmission_mediums)


def get_iterm_graphics_info():
    program = os.environ.get('TERM_PROGRAM')
    is_supported = False

    if program is not None:
        if 'iTerm' in program or 'WezTerm' in program or 'mintty' in program:
            is_supported = True

    return ITermGraphicsInfo(is_supported)


def get_info():
    return Info(get_size(), get_graphics_info())


def get_preferred_graphics_protocol():
    graphics_info = get_graphics_info()

    if graphics_info.kitty.is_supported:
        return 'kitty'
    elif graphics_info.iterm.is_supported:
        return 'iterm'
    elif graphics_info.sixel.is_supported:
        return 'sixel'
    else:
        return 'text'



#!/usr/bin/env python

import re

import numpy
from setuptools import Extension
from setuptools import find_packages
from setuptools import setup


def find_version():
    return re.search(r"^__version__ = '(.*)'$",
                     open('terminal_graphics/version.py', 'r').read(),
                     re.MULTILINE).group(1)

setup(name='terminal_graphics',
      version=find_version(),
      description=('Images in the terminal.'),
      long_description=open('README.rst', 'r').read(),
      author='Erik Moqvist',
      author_email='erik.moqvist@gmail.com',
      license='MIT',
      keywords=['terminal', 'image'],
      url='https://github.com/eerimoq/terminal_graphics',
      packages=find_packages(exclude=['tests']),
      install_requires=[
          'Pillow',
          'numpy',
          'rich'
      ],
      ext_modules=[
          Extension(name="terminal_graphics.ctext",
                    sources=["terminal_graphics/ctext.c"],
                    include_dirs=[numpy.get_include()])
      ],
      python_requires='>=3.6',
      test_suite="tests",
      entry_points = {
          'console_scripts': ['terminal_graphics=terminal_graphics:_main']
      })



#
# https://iterm2.com/documentation-images.html
#

from base64 import b64encode


def write(data, fout, size=None):
    arguments = f'size={len(data)};inline=1'

    if size is not None:
        arguments += f';width={size[0]};height={size[1]}'

    fout.write(f'\033]1337;File={arguments}:'.encode('utf-8'))
    fout.write(b64encode(data))
    fout.write(b'\a')
